#!/usr/bin/env python
import re
import numpy
import sys, os
import subprocess
import numpy as np
from astropy.time import Time
from pyrap.tables import table
from astropy import units as u
from casacore.tables import table
from  MSUtils.msutils import addcol
from astropy.coordinates import Angle
from astropy.coordinates import SkyCoord
from astropy.coordinates import solar_system_ephemeris, EarthLocation, AltAz
from astropy.coordinates import get_body_barycentric, get_body, get_moon


#________________________________________________________________________________________________________________________________________________________

#Auxiliar functions

def hms2deg(hms):
    '''
    Function to convert hms to degrees
    '''
    hms_angle = Angle(hms, unit='hour')
    return hms_angle.degree


def dms2deg(dms):
    '''
    Function to convert dms to degrees
    '''
    dms_angle = Angle(dms, unit='degree')
    return dms_angle.degree


def extract_scan_number(ms_scan):
    # Extract the scan number from the scan file name using re
    scan_number = re.search(r"scan_(\d+)\.ms", ms_scan).group(1)
    return int(scan_number)

#________________________________________________________________________________________________________________________________________________________


def extract_and_save_scan_numbers(ms, output_file):
    '''
    Extracts unique scan numbers from the Measurement Set
    measurement_set (str): Measurement Set path
    output_file (str): Output file path to save the scan numbers
    Returns:
    List[str]: List of unique scan numbers as strings
    '''
    print(f"Extracting scan numbers from {ms}...")
    
    scans = []
    
    with table(ms, ack=False) as maintab:
        scan_numbers = list(numpy.unique(maintab.getcol('SCAN_NUMBER')))
        scans = [str(scan) for scan in scan_numbers]

    print("Unique scan numbers extracted:")
    print(scans)

    # Save scan numbers to the output file
    with open(output_file, 'w') as file:
        file.write('\n'.join(scans))

    print(f"Scan numbers saved to {output_file}")
    
    return scans


#________________________________________________________________________________________________________________________________________________________

def rename_model_data_column(ms, oldname, newname):
    
    '''
    Rename the 'MODEL_DATA' column to 'MODEL_DATA_ORIGINAL' in a Measurement Set (MS) table.
    Parameters:
    ms (str): Path to the Measurement Set file.
    oldname (str): Name of the column to be renamed.
    newname (str): New name for the column.
    '''
    print(f"Processing {ms}...")
    # open the MS table in read-write mode
    ms = table(ms, readonly=False)
    # rename the 'MODEL_DATA' column to 'MODEL_DATA_ORIGINAL'
    ms.renamecol(oldname, newname)
    print(f"Column '{oldname}' renamed to '{newname}' in {ms}.")
    # close the MS table
    ms.close()

##########################################################################################################################################################

def rename_columns(ms_list, oldname, newname):
    '''
    Rename columns in a list of MS tables.
    '''
    for ms_name in ms_list:

        print(f"Renaming column '{oldname}' to '{newname}' in {ms_name}...")

        '''
        Open the MS table in read-write mode
        '''
        ms = table(ms_name, readonly=False)

         # Rename the 'oldname' column to 'newname'
        if 'oldname' in ms.colnames():
            ms.renamecol('oldname', 'newname')
        # close the MS table
        ms.close()
        print(f"Renaming completed for {ms_name}.")
    print("Rename column completed successfully.")

#________________________________________________________________________________________________________________________________________________________

def get_old_coords(ms_list, output_file):
    '''
    Extracts the old coordinates (ra and dec) of the foV from a measurement set using the PHASE_DIR column and writes them to a file.

    Parameters:
    ms_list (str): Path to the measurement sets files.
    outfile (str): Path to the output file.

    Returns:
    None
    '''
    sorted_ms_list = sorted(ms_list, key=lambda x: int(x.split('_scan_')[1].split('.')[0]))

    old_coords = []
    for ms in sorted_ms_list:

        print(f"Processing {ms}...")
        # 1. Read the PHASE_DIR column, the values in radians
        field_dir = table(f"{ms}::FIELD", readonly=True)
        phase_dir = field_dir.getcol("PHASE_DIR")

        # Extract the RA and Dec values from the PHASE_DIR column
        ra = phase_dir[0,0,0]
        dec = phase_dir[0,0,1]

        #2.convert to sky coordinates hms/dms
        sk = SkyCoord(ra*u.rad, dec*u.rad)

        # format th ra to hms
        ra_hms = sk.ra.to_string(unit=u.hourangle, pad=True)#, precision=1)

        # format dec to dms
        dec_dms = sk.dec.to_string(unit=u.degree, pad=True, alwayssign= True)#,precision=1)
        old_coords.append(f"{ra_hms} {dec_dms}")
        field_dir.close()
        print(f"Processing of {ms} complete.")

    with open(output_file,'wt') as f:
        for old_coord in old_coords:
            f.write(old_coord)
            f.writelines('\n')
    print("Old coordinates extraction complete.")
    return old_coords

#________________________________________________________________________________________________________________________________________________________

def get_sun_coordinates(ms, output_file):
    '''
    Extracts the coordinates of the Sun from a measurement set and writes them to a file.

    Parameters:
    ms (str): Path to the measurement set file.
    outfile (str): Path to the output file.

    Returns:
    None
    '''
    def format_coords(ra0,dec0):
        c = SkyCoord(ra0*u.deg,dec0*u.deg,frame='fk5')
        hms = str(c.ra.to_string(u.hour))
        dms = str(c.dec)
        return hms,dms

    # MeerKAT
    obs_lat = -30.71323598930457
    obs_lon = 21.443001467965008
    loc = EarthLocation.from_geodetic(obs_lat, obs_lon)

    maintab = table(ms, ack=False)
    scans = list(numpy.unique(maintab.getcol('SCAN_NUMBER')))
    lines = []

    print(f"Extracting Sun coordinates from {ms}...")
    for scan in scans:
        subtab = maintab.query(query='SCAN_NUMBER==' + str(scan))
        t_scan = numpy.mean(subtab.getcol('TIME'))
        t = Time(t_scan / 86400.0, format='mjd')

        with solar_system_ephemeris.set('builtin'):
            sun = get_body('Sun', t, loc)
            sun_ra = sun.ra.value
            sun_dec = sun.dec.value
            sun_hms, sun_dms = format_coords(sun_ra, sun_dec)
            lines.append(f"{sun_hms} {sun_dms}")

    maintab.close()

    with open(output_file, 'wt') as f:
        for line in lines:
            f.write(line + '\n')

    print(f"Sun coordinates extracted and saved to {output_file}.")


#________________________________________________________________________________________________________________________________________________________


def shift_coordinates(ms_list, coords, splitted_ms_dir):
    '''
    A funtion that takes a list of scans and coordinates shift/rephase it for a specific colunm (CORRECTED_DATA) and iutput in the splitted_ms_dir directory 
    Parameters:
    ms_list (list): Path to the Measurement Sets.
    coords (File): Path to the coordinate file.
    splitted_ms_dir (Directory): Path to the scans directory
    datacolumn (str): Datacolumn to use (when not defined default is 'all').
    '''
    with open(coords) as f:
        lines = f.readlines()
        num_lines = len(lines)

    #Sort the ms_list in numerical order
    sorted_ms_list = sorted(ms_list, key=lambda x: int(x.split('_scan_')[1].split('.')[0]))

    for i, ms in enumerate(sorted_ms_list):
        splitted_ms = os.path.join(splitted_ms_dir, os.path.basename(ms).replace(".ms", ".ms"))
        line = lines[i % num_lines].strip()
        print(f"Changing phase centre coordinates for {ms}...")
        os.system(f"chgcentre {ms} {line} {splitted_ms}")
        print(f'Scan {ms} Done.')
    print("Phase center changed successfully.")


#________________________________________________________________________________________________________________________________________________________

def create_ds9_region_from_file(input_file, output_dir, ms):
    """
    Function to create a DS9 region file for each coordinate in the input file
    """
    #Open the MS table
    tab = table(ms, readonly=True)
    # Extract the scan numbers
    scan_numbers = list(np.unique(tab.getcol('SCAN_NUMBER')))
    # Close the MS table
    tab.close()

    with open(input_file, 'r') as f:
        for i, line in enumerate(f):
            ra, dec = map(str.strip, line.split())
            print('RA in hms format:', ra, 'DEC in dms format:', dec, output_dir)
            print(f'Processing coordinates for scan {scan_numbers[i]}...')

            ra_deg = hms2deg(ra)
            dec_deg = dms2deg(dec)
            print('RA in degrees:', ra_deg, 'DEC in degrees:', dec_deg)
            print(f'Creating DS9 region for scan {scan_numbers[i]}...')

            scan_number = scan_numbers[i]  # Use the scan number at the corresponding index
            print(f"Creating region for scan {scan_number}"  '..................................................................................')
            create_ds9_region(ra_deg, dec_deg, scan_number, output_dir)
    print("DS9 region creation completed successfully.")

def create_ds9_region(ra, dec, scan_number, output_dir):
    # Function to create a DS99 region file
    sun_region = f'''# Region file format: DS9 CARTA 3.0.0-beta.3 
global color=green dashlist=8 3 width=3 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1
fk5
circle({ra}, {dec}, 1188.0000") # color=green
'''
#The UHF band images of the sun has a radius of 1062.9118 and the L bands 1188.0000"
    with open(f'{output_dir}/sun_region_{scan_number}.reg', 'w') as f:
        f.write(sun_region)

#_____________________________________________________________________________________________________________________________________________________


def add_column_to_ms(ms, col_names, like_col):
    """
    Add columns to a single measurement file
    """
    success = False
    try:
        print(f"Opening {ms}...")
        tb = table(ms, readonly=False)
    except Exception as e:
        print(f"Error: {e}")
        return success
    
    for col_name in col_names:
        if col_name not in tb.colnames():
            """
            Get column description from column 'like_col'
            """
            desc = tb.getcoldesc(like_col)
            desc[str('name')] = str(col_name)
            desc[str('comment')] = str(desc['comment'].replace(" ", "_"))
            dminfo = tb.getdminfo(like_col)
            dminfo[str("NAME")] = "{}-{}".format(dminfo["NAME"], col_name) 
            print(f"Adding column '{col_name}' to {ms}...")       
            tb.addcols(desc, dminfo)
            success = True
    print(f"Columns added to {ms} successfully.")
    tb.close()
    return success


#________________________________________________________________________________________________________________________________________________________


#Copying Model data into the model-data-sun in the  ooriginal MS

def copy_model_data_to_model_data_sun(ms, ms_list, copycol, tocol):
    print(f"Copying {copycol} to {tocol} in {ms}...")

    # Open the main MS table as maintab
    maintab = table(ms, readonly=False)

    # Get the unique scan numbers from the SCAN_NUMBER column in maintab
    scans = list(np.unique(maintab.getcol('SCAN_NUMBER')))

    for scan_number in scans:

        # Find the corresponding MS scan file for the current scan number
        ms_scan = next((ms_scan for ms_scan in ms_list if extract_scan_number(ms_scan) == scan_number), None)

        if ms_scan:
            # Open the MS scan as the source_subtab
            source_subtab = table(ms_scan, readonly=True)

            # Query the target_subtab to get the rows for the current scan number
            target_rows = maintab.query(query='SCAN_NUMBER==' + str(scan_number))

            # Get the MODEL_DATA from the source_subtab
            source_model_data = source_subtab.getcol(copycol)

            # Get the shape of the source_model_data
            source_shape = source_model_data.shape

            # Check if the target_subtab has the tocol column
            if tocol not in target_rows.colnames():
                # Add the tocol column to the target_subtab with the source_shape
                target_rows.addcols(
                    columns={tocol: {'datatype': 'complex', 'shape': source_shape}}
                )

            # Get the MODEL_DATA_SUN from the target_subtab
            target_model_data_sun = target_rows.getcol(tocol)

            # Update the target_model_data_sun with the source_model_data
            target_model_data_sun[:] = source_model_data

            # Put the updated MODEL_DATA_SUN back into the target_subtab
            target_rows.putcol(tocol, target_model_data_sun)

            # Flush the changes to disk
            maintab.flush()

            # Close the source_subtab
            source_subtab.close()
            print(f"Copy completed for scan {scan_number}.")

    # Close the maintab
    maintab.close()

    print(f"Copying {copycol} to {tocol} in {ms} completed successfully.")


#________________________________________________________________________________________________________________________________________________________




from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='solarkat',  
    version='1.0.3', 
    author='Victória Samboco',
    author_email='vicky.samboco@gmail.com', 
    description='Solar imaging pipeline for solar interference mitigation in MeerkAT', 
    repository='https://github.com/ratt-ru/solarkat.git',  
    documentation = 'https://solarkat-docs.readthedocs.io/en/latest/',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'aiobotocore==2.11.1',
        'aiohttp==3.9.3',
        'aioitertools==0.10.0',
        'aiosignal==1.2.0',
        'antlr4-python3-runtime==4.9.3',
        'appdirs==1.4.4',
        'asciitree==0.3.3',
        'astLib==0.11.10',
        'astro-kittens==1.4.6',
        'astro-tigger-lsm==1.7.2',
        'astropy==5.1',
        'async-timeout==4.0.2',
        'attrs==22.1.0',
        'bokeh==2.4.3',
        'botocore==1.27.59',
        'breizorro==0.1.0',
        'certifi==2022.9.24',
        'charset-normalizer==2.1.1',
        'cli-ui==0.17.2',
        'click==8.1.3',
        'cloudpickle==2.2.0',
        'codex-africanus==0.3.3',
        'colorama==0.4.5',
        'Columnar==1.4.1',
        'commonmark==0.9.1',
        'configparser==5.3.0',
        'contextlib2==21.6.0',
        'contourpy==1.0.5',
        'cycler==0.11.0',
        'dask==2022.9.2',
        'dask-ms==0.2.14',
        'decorator==5.1.1',
        'dill==0.3.5.1',
        'distributed==2022.9.2',
        'docopt==0.6.2',
        'donfig==0.7.0',
        'entrypoints==0.4',
        'fasteners==0.18',
        'fonttools==4.38.0',
        'frozenlist==1.3.1',
        'fsspec==2022.8.2',
        'future==0.18.2',
        'future-fstrings==1.2.0',
        'HeapDict==1.0.1',
        'idna==3.4',
        'importlib-metadata==5.0.0',
        'iniconfig==1.1.1',
        'Jinja2==3.1.2',
        'jmespath==1.0.1',
        'kiwisolver==1.4.4',
        'llvmlite==0.39.1',
        'locket==1.0.0',
        'loguru==0.6.0',
        'MarkupSafe==2.1.1',
        'matplotlib==3.6.1',
        'msgpack==1.0.4',
        'msutils==1.2.0',
        'multidict==6.0.2',
        'multiprocess==0.70.13',
        'munch==2.5.0',
        'nose==1.3.7',
        'numba==0.56.2',
        'numcodecs==0.10.2',
        'numpy==1.23.3',
        'omegaconf==2.2.3',
        'packaging==21.3',
        'pandas==1.5.0',
        'partd==1.3.0',
        'pathos==0.2.9',
        'Pillow==9.2.0',
        'pkg_resources==0.0.0',
        'pluggy==1.0.0',
        'pox==0.3.1',
        'ppft==1.7.6.5',
        'psutil==5.9.2',
        'py==1.11.0',
        'pydantic==1.10.2',
        'pyerfa==2.0.0.1',
        'Pygments==2.13.0',
        'pyparsing==3.0.9',
        'pytest==7.1.3',
        'python-casacore==3.5.2',
        'python-dateutil==2.8.2',
        'pytz==2022.4',
        'PyYAML==6.0.1',
        'quartical==0.2.1',
        'regions>=0.5',
        'requests==2.28.1',
        'rich==12.5.1',
        'ruamel.yaml==0.17.21',
        'ruamel.yaml.clib==0.2.6',
        's3fs==2022.8.2',
        'scabha==2.0.0',
        'schema==0.7.5',
        'scipy==1.9.1',
        'six==1.16.0',
        'sortedcontainers==2.4.0',
        'stimela2==2.0rc4',
        'tabulate==0.8.10',
        'tblib==1.7.0',
        'tbump==6.9.0',
        'tomli==2.0.1',
        'tomlkit==0.11.5',
        'toolz==0.12.0',
        'tornado==6.1',
        'typing_extensions==4.3.0',
        'Unidecode==1.3.6',
        'urllib3==1.26.12',
        'wcwidth==0.2.5',
        'wrapt==1.14.1',
        'xarray==2022.9.0',
        'yarl==1.8.1',
        'zarr==2.13.2',
        'zict==2.2.0',
        'zipp==3.8.1',
        'breizoirro==0.1.2',
    ],

    classifiers=[ 
        'Development Status :: 4 - Beta', 
        'Intended Audience :: Science/Research',  
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Astronomy',
    ],

    include = [
    'solarkat/solarkat-pipeline/solarkat.yaml',
    'solarkat/solarkat-pipeline/solarkat-cabs.yaml',
    ]
)